import json , uuid , os
from base64 import b64encode
from http.client import HTTPSConnection
import xmltodict , dicttoxml
from xml.dom.minidom import parseString
from datetime import datetime
import pyodbc
import re

from MedatechUK.oDataConfig import Config
from MedatechUK.mLog import mLog
from MedatechUK.apy import Response

#region "Form Class"
class SerialF :

    #region Properties
    @property
    def fname(self):    
        return self._fname
    @fname.setter
    def fname(self, value):
        self._fname = value

    @property
    def rt(self):    
        return str(self._rt)
    @rt.setter
    def rt(self, value):
        self._rt = value

    @property
    def typename(self):    
        return str(self._typename)
    @typename.setter
    def typename(self, value):
        self._typename = value

    @property
    def bubbleid(self):    
        return str(self._bubbleid)

    #endregion

    #region "ctor"
    def __init__(self, **kwargs):
        self.fname = ''
        self.rt = 1
        self._bubbleid = uuid.uuid4()
        self.typename = ""

        for arg in kwargs.keys():    
            if ( hasattr( self , arg ) ):
                setattr( self , arg , kwargs[arg] )   
    
    #endregion

#endregion

#region "Type Class"
class SerialT :
    
    #region Propereties
    @property
    def pType(self):    
        return self._pType
    @pType.setter
    def pType(self, value):
        self._pType = value

    @property
    def Len(self):    
        return self._Len
    @Len.setter
    def Len(self, value):
        self._Len = value

    @property
    def pCol(self):    
        return self._pCol
    @pCol.setter
    def pCol(self, value):
        self._pCol = value

    #endregion

    #region "ctor"
    def __init__(self, this, att , **kwargs):
        
        self.pCol = att
        self.pType = 'CHAR'
        self.Len = 40
        
        for arg in kwargs.keys():    
            if ( hasattr( self , arg ) ):
                setattr( self , arg , kwargs[arg] )              

        if (att=="rt"):
            self.pCol="RECORDTYPE" 
            self.pType="CHAR"
        
        if (att=="bubbleid"):
            self.pCol="BUBBLEID" 
            self.pType="CHAR"
        
        if (att=="typename"):
            self.pCol="TYPENAME" 
            self.pType="CHAR"
            self.len = 3

        this.props[att] = self
    
    #endregion

    #region "Methods"
    def oData(self, this):
        return {
            'CHAR': chr(34) + self.pCol + chr(34) + " : " + chr(34) + str(this) + chr(34) ,
            'INT': chr(34) + self.pCol + chr(34) + " : " + str(this) ,
            'REAL': chr(34) + self.pCol + chr(34) + " : " + str(this) ,
        }.get(self._pType.upper(), chr(34) + self.pCol + chr(34) + " : " + chr(34) + str(this) + chr(34))

    def SQLColumn(self, this):
        return self.pCol.upper()

    def SQLValue(self, this):
        return {
            'CHAR': chr(39) + str(this) + chr(39) ,
            'INT':  str(this) ,
            'REAL': str(this) ,
        }.get(self._pType.upper(), chr(39) + str(this) + chr(39))

    def XML(self, this):
        return "<" + self._pCol + ">" + str(this) + "</" + self._pCol + ">"

    #endregion

#endregion 

#region "Base Class"

class SerialBase :

    #region Properies
    @property
    def props(self):    
        return self._props
    @props.setter
    def props(self, value):
        self._props = value

    #endregion

    #region "ctor"
    def __init__(self, form , **kwargs):
        self.form = form
        self.props = {}
        self.log = mLog() 
        self.ret = {}  

        #region Create instance from input
        for arg in kwargs.keys():  
            if arg.upper() == '_XML':   
                self.log.logger.debug("Create instance from [{}] file {}".format(arg.upper().lstrip("_") , kwargs[arg].name))
                t = xmltodict.parse(
                    kwargs[arg].read(),
                    dict_constructor=dict
                )
                self = self.__init__(**json.loads(json.dumps(t[list(t)[0]]))) #

            elif arg.upper() == '_SQL':                                   
                self.log.logger.debug("Create instance from [{}] procedure {}".format(arg.upper().lstrip("_") , kwargs['_sql']['proc']))
                self = self.__init__(
                    **json.loads(
                        json.dumps(
                            self.fromSQL(
                                kwargs['_sql']['config'] , 
                                kwargs['_sql']['proc'] , 
                                **kwargs['_sql']['kwargs'] 
                            )                    
                        )
                    )
                ) #

            elif arg.upper() == '_JSON':
                self.log.logger.debug("Create instance from [{}] file {}".format(arg.upper().lstrip("_") , kwargs[arg].name))
                self = self.__init__(**json.loads(kwargs[arg].read()))
        
        #endregion

        #region Set Attributes
            elif ( hasattr( self , arg ) ) :
                try:
                    setattr( self , arg , kwargs[arg] )
                except:
                    pass
            elif ( hasattr( self , arg.lstrip('@') ) ) :
                try:
                    setattr( self , arg.lstrip('@') , kwargs[arg] )
                except e:
                    print(e)

        #endregion

    def fromSQL(self, config, endpoint, **kwargs):
        try:
            self.log.logger.debug("Opening DB query: [{}] in [{}].".format( endpoint , config.environment ))
            cnxn = pyodbc.connect("Driver={SQL Server Native Client 11.0};DATABASE=system;" + config.connstr )
            crsr = cnxn.cursor() 
            sql = ["select SO.OBJECT_ID as [ObjectID], " , 
                "SCHEMA_NAME(SCHEMA_ID) + '.' + SO.name AS [ObjectName] " ,
                "From sys.objects AS SO " ,
                "INNER JOIN sys.parameters AS P " ,
                "On SO.OBJECT_ID = P.OBJECT_ID " ,
                "WHERE 0=0 " ,
                "And SO.TYPE IN ('P') " ,
                "And (TYPE_NAME(P.user_type_id)='xml') " ,
                "And (LOWER(SO.name)=LOWER('"+ endpoint +"')) " ,
                "And P.is_output=1 "]
                
            crsr.execute(
                ('\n'.join(sql))
            )

            row = crsr.fetchone()       
            if row != None :                           
                sql = ["SELECT	" ,
                "	P.name AS [ParameterName],	" ,
                "	TYPE_NAME(P.user_type_id) AS [ParameterDataType] " ,
                "FROM sys.objects AS SO	" ,
                "	INNER JOIN sys.parameters AS P 	" ,
                "	ON SO.OBJECT_ID = P.OBJECT_ID	" ,
                "WHERE 0=0	" ,
                "	And SO.OBJECT_ID = "+ str(row.ObjectID) ,
                "	And P.is_output=0" ,
                "order by parameter_id"
                ]
                crsr.execute(    
                    '\n'.join(sql)
                )
                ObjectName = row.ObjectName
                par = []
                for row in crsr.fetchall() :
                    if not kwargs.__contains__(row.ParameterName[1:]) and not row.ParameterName[1:] == "ENV":
                        raise NameError("Missing parameter {}".format(row.ParameterName[1:]))
                    
                    if row.ParameterDataType in ["nchar", "nvarchar", "ntext"]:
                        if row.ParameterName[1:].upper() == "ENV":
                            par.append("N'{}'".format(config.environment))                            
                        else:
                            par.append("N'{}'".format(kwargs[(row.ParameterName[1:])]))
                    
                    elif row.ParameterDataType in ["char", "varchar", "text"]:
                        if row.ParameterName[1:].upper() == "ENV":
                            par.append("'{}'".format(config.environment))
                        else:                            
                            par.append("'{}'".format(kwargs[(row.ParameterName[1:])]))
                    
                    else:                      
                        par.append("{}".format( kwargs[(row.ParameterName[1:],"0")]))
                
                par.append("@Output_XML OUT ")

                sql = [
                    "DECLARE  @Output_XML xml ",
                    "EXEC {} {}".format(ObjectName , ", ".join(par)) ,                    
                    "SELECT @Output_XML"
                ]
                crsr.execute("\n".join(sql))
                
                row = crsr.fetchone()
                ret = xmltodict.parse(row[0])    
                return ret[list(xmltodict.parse(row[0]))[0]]

            else:
                raise NameError("Invalid SQL Procedure {}".format(endpoint))
        
        except Exception as e:
            raise e
        
    #endregion

    #region "Output Methods"

    def toFile(self, fn , method, **kwargs):
        self.log.logger.debug("Writing file [{}] {}".format( fn , method(**kwargs)))
        with open(fn, 'w') as the_file:            
            the_file.write(method(**kwargs)) 

    #region "Built in toFile() methods"

    def toXML(self , this = 0, root="root"):         
        ret = ""  
        l = 0     
        if (this==0):            
            l = 1
            this = self
            ret += "<{}>".format(root) 

        if isinstance(this, list):            
            for i in this:
                ret += "{}{}{}".format(
                    "<{}>".format( 
                        type(i).__name__ 
                    ) ,
                    self.toXML(i) ,
                    "</{}>".format( 
                        type(i).__name__ 
                    )
                )
        else:
            for p in range(3):
                for key in [key for key in this.__dict__ if key not in ["_props","form","log","ret"]]:  
                    if isinstance(this.__dict__[key], list) and p==2:      
                        ret += self.toXML(this.__dict__[key])

                    elif hasattr(getattr(this , key), "props") and p==1:
                        if(hasattr(this,"props")):                            
                            ret += "{}{}{}".format(
                                "<{}>".format( 
                                    key.lstrip('_') 
                                ) , 
                                self.toXML(this.__dict__[key]) , 
                                "</{}>".format( 
                                    key.lstrip('_') 
                                ) 
                            )
                        
                    elif p==0:
                        if(hasattr(this,"props")) and this.props.__contains__(key.lstrip('_')):
                            ret += "{}{}{}".format(
                                "<{}>".format(
                                    key.lstrip('_') 
                                ) ,
                                str(this.__dict__[key]) ,
                                "</{}>".format(
                                    key.lstrip('_')
                                )
                            )
                
        if(l!=0):
            ret += "</{}>".format( root ) 
            return parseString(ret).toprettyxml()            
        
        else:
            return ret

    def toJSON(self):        
        return json.dumps(self, default=lambda o: {
            key.lstrip('_'): 
            value for key, value in o.__dict__.items() if not key in ["_props" , "form" , "log" , "ret"]
            }, 
                sort_keys=False, 
                indent=4
        )   

    #endregion
  
    def toPri(self, config, method, **kwargs):            
        
        ## This function will PATCH a completion request ONLY
        ## if the fornname of the upper level is 'ZODA_TRANS'

        # Set the response object
        if kwargs.__contains__("request"):
            # The call contains a request - populate it's response
            ret = kwargs["request"].response
        elif kwargs.__contains__("response"):
            # The call contains a response - populate directly
            ret = kwargs["response"]
            
        url = '/odata/priority/{}/{}/{}'.format(config.tabulaini , config.environment , "ZODA_TRANS")
        headers= { 
                'Authorization' : 'Basic %s' %  b64encode(bytearray(config.ouser + ":" + config.opass,'ascii')).decode("ascii") ,
                'Content-Type': 'application/json',
                "User-Agent": "MedatechUK Python Client",
            }
        data = json.loads(method())
        self.log.logger.debug("POSTing to [{}{}] ".format( config.oDataHost, url ))                 
        self.log.logger.debug("Headers:\n{}".format( json.dumps(headers ,  indent = 4) ))
        self.log.logger.debug("Data:\n{}".format( json.dumps(data, indent = 4) ))        

        r = HTTPSConnection(config.oDataHost)  
        r.request( 
            'POST', 
            url , 
            json.dumps(data),
            headers,              
        )
        res = r.getresponse()             
        if res.status == 201: # Created
            self.log.logger.debug("[{}] OK".format( res.status ))

            # If we're using the oData loading form, send a PATCH
            # to identify that all data has been sent.
            if self.form.fname != 'ZODA_TRANS':
                self.log.logger.debug("[{}] {}".format( res.status , res.reason ))
                ret.Status = res.status
                ret.Message = res.reason
                ret.data = json.load(res)
                self.log.logger.debug("Result: {}".format( json.dumps(ret.data  , indent = 4 )))

            else:
                data = json.loads(res.read())    

                patch = {}
                patch['COMPLETE'] = "Y"

                r.request( 
                    'PATCH', 
                    url + "(BUBBLEID='"+ data['BUBBLEID'] + "',LOADTYPE=" + str(data['LOADTYPE']) + ")", 
                    headers=headers, 
                    body=json.dumps(patch) 
                )
                
                self.log.logger.debug("PATCHing to [{}] ... ".format( url + "(BUBBLEID='"+ data['BUBBLEID'] + "',LOADTYPE=" + str(data['LOADTYPE']) + ")" ))  
                res = r.getresponse()

                if res.status != 200: # PATCHed
                    self.log.logger.critical("[{}] Fail: {}".format( res.status , res.reason ))
                    ret.Status = res.status   
                    ret.Message = "PATCH Failed: " + res.reason  
                    
                    # If the response is text, create a response with the text         
                    if res.getheader("Content-Type","").find("text/plain") > -1:                             
                        er = str(res.read().decode('utf-8'))
                        ret.data = {"error": er }     
                        self.log.logger.critical("{}".format( er ))              

                    elif res.getheader("Content-Type","").find("text/html") > -1:
                            ret.data = {"error": "Priority service not responding." }     
                            self.log.logger.critical("{}".format( "Priority service not responding." ))   

                    else:
                        # Create reponse from json 
                        ret.data = json.load(res)  
                        self.log.logger.critical( "{}".format( json.dumps(ret.data , indent = 4 ) ) )

                else:
                    ## Sucsess!
                    self.log.logger.debug("[{}] {}".format( res.status , res.reason ))
                    ret.Status = res.status
                    ret.Message = res.reason
                    ret.data = json.load(res)
                    self.log.logger.debug("Result: {}".format( json.dumps(ret.data , indent = 4 ) ))

        else:   
            ret.Status = res.status
            ret.Message = "POST Failed: " + res.reason   
            self.log.logger.critical( "[{}] Fail: {}".format( res.status , res.reason ) )            

            if len(ret.data)!=0:
                # If the response is text, create a response with the text         
                if res.getheader("Content-Type","").find("text/plain") > -1:                             
                    er = str(res.read().decode('utf-8'))
                    ret.data = {"error": er }     
                    self.log.logger.critical("{}".format( er ))              

                elif res.getheader("Content-Type","").find("text/html") > -1:
                        ret.data = {"error": "Priority service not responding." }     
                        self.log.logger.critical("{}".format( "Priority service not responding." ))   

                elif res.getheader("Content-Type","").find("json") > -1:
                    # Create reponse from json 
                    ret.data = json.load(res)  
                    self.log.logger.critical( "{}".format( json.dumps(ret.data  , indent = 4 ) ) )
    
    #region "Built in toPri() oData methods"

    def toOdata(self , this = 0):         
        ret = ""  
        l = 0     
        if (this==0):            
            l = 1
            this = self
            ret +=("{ ") 

        if isinstance(this, list):
            ret +=("[ ")
            for i in this:
                ret +=("{ ")
                ret +=self.toOdata(i)
                ret +=(" }")
                if(this[-1]!=i):
                    ret += (" , ")
            ret +=(" ]")
            
        else:
            f = 0
            for key in this.__dict__ :  
                if (key != "_props" and key !="form"and key !="log"):
                    if isinstance(this.__dict__[key], list):      
                        ret += ' , ' +chr(34)+ (this.__dict__[key][-1].form.fname) +chr(34)+ ' : '
                        ret += self.toOdata(this.__dict__[key])
                    
                    else:
                        if(hasattr(this,"props")) :
                            if(f==0):
                                f = 1
                            else :
                                ret +=(", ")
                            ret +=(this.props[key.lstrip('_')].oData(this.__dict__[key]))        

            # Iterate through readonly properties
            for key in this.props:
                if (key != "rt" and key !="bubbleid" and key !="typename"):
                    if not this.__dict__.__contains__("_" + key):
                        if(f==0):
                            f = 1
                        else :
                            ret +=(", ")
                        ret +=(this.props[key.lstrip('_')].oData(getattr(this, key)))   

        if(l!=0):
            ret += " }"

        return ret

    def toFlatOdata(self , this = 0): 
    
        ## Returns the object as oData commands for flat loading     
        ret = ""  
        l = 0     
        if (this==0):            
            l = 1
            this = self
            ret +=("{ ")             
            ret +=(this.props["bubbleid"].oData(this.form.bubbleid) )  + ", "            
            ret +=(this.props["typename"].oData(this.form.typename) )  + ", " + chr(34) + "ZODA_LOAD_SUBFORM" + chr(34) + " : ["            

        if isinstance(this, list):            
            for i in this:                               
                ret += self.toFlatOdata(i)                              
                if(this[-1]!=i) and ret[len(ret)-4:] != " } ,":
                    ret += (" } ,")

        else:
            f = 0
            ret += " { "
            if(this.props.__contains__("rt")) :
                if(f==0): 
                    f = 1
                else :
                    ret +=(", ")
                ret +=(this.props["rt"].oData(this.form.rt) ) 
            
            # Iterate through readonly properties
            for key in [key for key in this.props if key not in ["rt","bubbleid","bubbleid","typename"]]:
                if not this.__dict__.__contains__("_" + key):
                    if(f==0):
                        f = 1
                    else :
                        ret +=(", ")
                    ret +=(this.props[key.lstrip('_')].oData(getattr(this, key))) 

            for p in range(3):
                for key in [key for key in this.__dict__ if key not in ["_props","form","log","ret"]]:                
                    if isinstance(this.__dict__[key], list) and p==2:      
                        if ret[len(ret)-4:] != " } ,":
                            ret+=" } ,"
                        ret += self.toFlatOdata(this.__dict__[key])                            

                    elif hasattr(getattr(this , key), "props") and p==1:
                        if ret[len(ret)-4:] != " } ,":
                            ret+=" } ,"
                        ret += self.toFlatOdata(this.__dict__[key])
                        
                    elif p==0:
                        if(hasattr(this,"props")) and this.props.__contains__(key.lstrip('_')):
                            if(f==0):
                                f = 1
                            else :
                                ret +=(", ")
                            ret +=(this.props[key.lstrip('_')].oData(this.__dict__[key]))                                                 
            
        if(l!=0):
            if ret[len(ret)-4:] == " } ,":
                ret=ret[0:len(ret)-4]  
            
            ret += " } ] }"

        return ret    

    #endregion

    def toSQL(self , this = 0):        

        ## Returns the object as SQL commands 
         
        names = {}
        l = 0     
        if (this==0):     
            self.ret = {}       
            l = 1
            this = self            
        
            names[len(names)] = ({"name" : "BUBBLEID", "value" : this.props["bubbleid"].SQLValue(this.form.bubbleid)})
            names[len(names)] = (
                {
                    "name" : "LOADTYPE", 
                    "value" : "( select TYPE from ZODAT_TYPE where TYPENAME = {} )".format(
                        this.props["typename"].SQLValue(this.form.typename)
                    )
                }
            )
            names[len(names)] = (
                {
                    "name" : "LINE", 
                    "value" : "( SELECT MAX(LINE)+1 FROM ZODAT_TRANS )"
                }
            )
            
            self.ret[len(self.ret)] = names
            self.ret[len(self.ret)-1]["TABLE"] = this.form.fname
            names = {}
        
        names[len(names)] = (
            {
                "name" : "PARENT", 
                "value" : "( SELECT LINE FROM ZODAT_TRANS WHERE BUBBLEID = {})".format(
                    self.ret[0][0].get("value")
                )                 
            }
        )
        names[len(names)] = (
            {
                "name" : "LINE", 
                "value" : "{}".format(str(len(self.ret)))
            }
        )        

        if isinstance(this, list):            
            for i in this:      
                if len(names) > 2:  
                    self.ret[len(self.ret)] = names 
                    self.ret[len(self.ret)-1]["TABLE"] = i.form.fname   
                    names = {}                   
                self.toSQL(i)                              

        else:                        
            names[len(names)] = ({"name" : "RECORDTYPE", "value" : this.props["rt"].SQLValue(this.form.rt)})                 
            
            # Iterate through readonly properties
            for key in this.props:
                if (key != "rt" and key !="bubbleid" and key !="typename"):
                    if not this.__dict__.__contains__("_" + key):                                              
                        names[len(names)] = (
                            {
                                "name" : this.props[key.lstrip('_')].SQLColumn(getattr(this, key)), 
                                "value" : this.props[key.lstrip('_')].SQLValue(getattr(this, key)) 
                            }
                        )        

            for p in range(3):
                for key in this.__dict__ :                      
                    if (key != "_props" and key !="form" and key !="log"):
                        # print("{} {}".format(p,key))

                        if isinstance(this.__dict__[key], list) and p==2: 
                            if len(names) > 2:  
                                self.ret[len(self.ret)] = names
                                self.ret[len(self.ret)-1]["TABLE"] = this.form.fname
                                names = {}
                            self.toSQL(this.__dict__[key])                   

                        elif hasattr(getattr(this , key), "props") and p==1:
                            if len(names) > 2:
                                self.ret[len(self.ret)] = names
                                self.ret[len(self.ret)-1]["TABLE"] = this.form.fornname
                                names = {}
                            self.toSQL(this.__dict__[key])  
                            
                        elif p==0:
                            if(hasattr(this,"props")) and this.props.__contains__(key.lstrip('_')):
                                names[len(names)] = (
                                    {
                                        "name" : this.props[key.lstrip('_')].SQLColumn(this.__dict__[key]), 
                                        "value" : this.props[key.lstrip('_')].SQLValue(this.__dict__[key]) 
                                    }
                                )                                                      

            if len(names) > 2:
                self.ret[len(self.ret)] = names
                self.ret[len(self.ret)-1]["TABLE"] = this.form.fname
                names = {}

        if (l!=0):   
            if len(self.ret) >2 :
                self.ret[1]["TABLE"] = self.ret[2]["TABLE"]

            ins = ""
            ins += "if exists(select TYPE from ZODAT_TYPE where TYPENAME = {})\nbegin\n".format(this.props["typename"].SQLValue(this.form.typename))
            ins += "\tset identity_insert ZODAT_TRANS  on\n" 
            for e in range(len(self.ret)) :                
                ins += "\tINSERT INTO {} ( ".format(self.ret[e]["TABLE"])                 
                
                for f in range(len(self.ret[e])-1) : 
                    try:
                        a = self.ret[e][f]["name"]
                        if f==0:
                            ins += "{}".format( self.ret[e][f].get("name") ) 
                        else:
                            ins += ", {}".format( self.ret[e][f].get("name") ) 
                    except:
                        pass

                ins += " ) VALUES ( "                
                for f in range(len(self.ret[e])-1) : 
                    try:
                        a = self.ret[e][f]["name"]
                        if f==0:
                            ins += "{}".format( self.ret[e][f].get("value") ) 
                        else:
                            ins += ", {}".format( self.ret[e][f].get("value") ) 
                    except:
                        pass               

                ins += " )\n" 
            
            d = datetime.now()
            ins += "\tUPDATE ZODAT_TRANS SET COMPLETE = 'Y' , COMPLETEDATE = {} WHERE BUBBLEID = {}\n".format(                              
                int(
                    (datetime(
                        d.year, 
                        d.month, 
                        d.day, 
                        d.hour, 
                        d.minute) 
                    - datetime(1988, 1, 1)).total_seconds() / 60
                ) ,
                self.ret[0][0].get("value")
            )
            ins += "\tset identity_insert ZODAT_TRANS  off\n"
            ins += "end"
            return ins

    #region "Built in toSQL() methods"

    def toPriSQL(self, config, method, **kwargs):

        # Set the response object
        if kwargs.__contains__("request"):
            # The call contains a request - populate it's response
            ret = kwargs["request"].response
        elif kwargs.__contains__("response"):
            # The call contains a response - populate directly
            ret = kwargs["response"]
                 
        # Environment Check
        if not config.CheckEnviroment():
            self.log.logger.critical("Bad environment [{}].".format( config.environment ))
            ret.Status = 400
            ret.Message = "Invalid company."
            ret.data = {"error" : "Company [" + config.environment + "] not found."}          

        else:
            ## Environment OK            
            try:
                cnxn = config.cnxn()

            except Exception as e:                
                ret.Status = 500
                ret.Message = "SQL Error" 
                ret.data = {"error" : str(e) }          
            
            try:
                crsr = cnxn.cursor() 
                self.log.logger.debug("use {}; {}".format(config.environment , method()))
                crsr.execute(
                    "use {}; {}".format(config.environment , method())
                )                         
                
                ret.Status = 200
                ret.Message = "OK"
                ret.data = {"error" : "No errors found" }          

            except Exception as e:
                self.log.logger.critical("TSQL Error: {}".format( str(e) ))
                ret.Status = 400
                ret.Message = "SQL Error" 
                ret.data = {"error" : str(e) }  

    #endregion      
    
    def toURL(self, url, method, **kwargs): 

        # Set the response object
        if kwargs.__contains__("request"):
            # The call contains a request - populate it's response
            ret = kwargs["request"].response
        elif kwargs.__contains__("response"):
            # The call contains a response - populate directly
            ret = kwargs["response"]
        else:
            ret = Response()

        verb = "POST"
        if kwargs.__contains__("method"):
            verb = kwargs("method")

        host = url.split("://")[1].split("/")[0]
        url = url.split(host)[1]

        headers = {}
        headers["User-Agent"] = "MedatechUK Python Client"
        if not kwargs.__contains__("Content-Type"):
            headers["Content-Type"] = "application/json"
        else:
            headers["Content-Type"] = kwargs["Content-Type"]

        if kwargs.__contains__("user") and kwargs.__contains__("passw"):
            headers["Authorization"] = 'Basic %s' %  b64encode(bytearray(kwargs["user"] + ":" + kwargs["passw"],'ascii')).decode("ascii")

        self.log.logger.debug("POSTing to [{}{}] ".format( host, url ))                 
        self.log.logger.debug("Headers:\n{}".format( json.dumps(headers ,  indent = 4) ))
        self.log.logger.debug("Data:\n{}".format( method() ))   

        r = HTTPSConnection(host)  
        r.request( 
            verb.upper(), 
            url , 
            method(),
            headers,              
        )
        res = r.getresponse()       

        if str(res.status)[0] == "2": # OK
            self.log.logger.debug("[{}] {}".format( res.status , res.reason ))
            ret.Status = res.status
            ret.Message = res.reason
            ret.data = json.load(res)
            self.log.logger.debug("Result: {}".format( json.dumps(ret.data  , indent = 4 )))
            
        else:   
            ret.Status = res.status
            ret.Message = "POST Failed: " + res.reason   
            self.log.logger.critical( "[{}] Fail: {}".format( res.status , res.reason ) )            

            if len(ret.data)!=0:
                # If the response is text, create a response with the text         
                if res.getheader("Content-Type","").find("text/plain") > -1:                             
                    er = str(res.read().decode('utf-8'))
                    ret.data = {"error": er }     
                    self.log.logger.critical("{}".format( er ))              

                elif res.getheader("Content-Type","").find("text/html") > -1:
                        ret.data = {"error": "Priority service not responding." }     
                        self.log.logger.critical("{}".format( "Priority service not responding." ))   

                elif res.getheader("Content-Type","").find("json") > -1:
                    # Create reponse from json 
                    ret.data = json.load(res)  
                    self.log.logger.critical( "{}".format( json.dumps(ret.data  , indent = 4 ) ) )
    
    #endregion

#endregion

#region infer object from file
class infer():

    #region "Ctor"
    def __init__(self, data, **kwargs):

        rt=0
        self.mycls = {}
        self.mylen = {}
        self.cls = []        
        self.imp = []
        self.preq = []

        self.recurse(data, kwargs['name'])

        self.imp.append("import json")
        self.imp.append("from MedatechUK.Serial import SerialBase , SerialT , SerialF")			
        self.imp.append("from MedatechUK.mLog import mLog")

        for k in self.mycls.keys():
            rt += 1
            baseclass = ("class {}(SerialBase) :\n".format(k))		
            props = []
            ctr = []
            met = []

            if rt==1:
                met.append("    SerialT(self, \"rt\")")
                met.append("    SerialT(self, \"bubbleid\")")
                met.append("    SerialT(self, \"typename\")")
                baseinit = "    SerialBase.__init__(self , SerialF(fname=\"ZODA_TRANS\", rt={}, typename=\"ORD\"), **kwargs)  ".format(rt)

                self.preq = [                    
                    "def ProcessRequest(request) :" ,
                    "    log = mLog()" ,
                    "    try:" ,
                    "        q = {}(**request.data) ".format(k) ,
                    "        q.toPri(" ,
                    "            Config(" ,
                    "                env=request.environment , " ,
                    "                path=os.getcwd()" ,
                    "            ) , " ,
                    "            q.toFlatOdata, " ,
                    "            request=request ",
                    "        )        ",
                    "    ",
                    "    except Exception as e:",
                    "        log.logger.critical(str(e))",
                    "        exc_type, exc_obj, exc_tb = sys.exc_info()",
                    "        fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]",
                    "        request.response.Status = 500" ,
                    "        request.response.Message = \"Internal Server Error\"" ,
                    "        request.response.data ={ \"error\" :",
                    "            {",
                    "                \"type\": exc_type,",
                    "                \"message\": str(e),",
                    "                \"script\": fname,",
                    "                \"line\": exc_tb.tb_lineno",
                    "            }",
                    "        } ",
                    ""
                ]

            else:
                met.append("    SerialT(self, \"rt\")")
                baseinit = "    SerialBase.__init__(self , SerialF(fname=\"ZODA_LOAD\", rt={}), **kwargs)  ".format(rt)
            
            xi = 0
            xr = 0 
            xt = 0

            for e, j in enumerate(self.mycls[k],1):

                if isinstance(self.mycls[k][j], list):
                    prop = [
                        "    ",
                        "@property" ,
                        "def {}(self):".format(j) ,
                        "    return self._{}".format(j) ,
                        "@{}.setter".format(j) ,
                        "def {}(self, value):".format(j) , 
                        "    self._{} = [] ".format(j) ,
                        "    if isinstance(value, list):",
                        "        for i in range(len(value)):",
                        "            self._{}.append({}(**value[i]))".format(j,j),
                        "    else:",
                        "        self._{}.append({}(**value))".format(j,j),
                        ""
                    ]                
                    ctr.append("    self.{} = []".format(j))

                elif self.mycls[k][j][0] == "$":
                    prop = [
                        "    ",
                        "@property" ,
                        "def {}(self):".format(j) ,
                        "   return self._{} ".format(j) ,
                        "@{}.setter".format(j),
                        "def {}(self, value):".format(j),
                        "   self._{} = {}(**value)".format(j,j),
                        ""
                    ]
                    ctr.append("    self.{} = {}{}".format(j , "{" , "}"))				

                else:
                    prop = [
                        "    ",
                        "@property" ,
                        "def {}(self):".format(j) ,
                        "   return self._{} ".format(j) ,
                        "@{}.setter".format(j),
                        "def {}(self, value):".format(j),
                        "   self._{} = value".format(j),
                        ""
                    ]				
                    if self.mycls[k][j] == 'INT':
                        xi+=1
                        met.append("    SerialT(self, \"{}\" , pCol=\"{}{}\" , pType=\"INT\")".format(j,self.mycls[k][j],xi))
                        ctr.append("    self.{} = 0".format(j))

                    elif self.mycls[k][j] == 'REAL':
                        xr+=1
                        met.append("    SerialT(self, \"{}\" , pCol=\"{}{}\" , pType=\"REAL\")".format(j,self.mycls[k][j],xr))
                        ctr.append("    self.{} = 0.0".format(j))

                    else :
                        xt +=1
                        met.append("    SerialT(self, \"{}\" , pCol=\"{}{}\" , Len={} , pType=\"CHAR\")".format(j,self.mycls[k][j],xt,self.mylen[k][j]))
                        ctr.append("    self.{} = \"\"".format(j))

                props.append("\n    ".join(prop))

            ctor = [
                "    #region \"ctor\"",
                "def __init__(self,  **kwargs): ",
                "",
                "    #region \"Property defaults\"" ,
                "{}".format("\n    ".join(ctr)), 
                "",
                "    #endregion",
                "",                    
                "    #region \"Set Meta info\"",
                baseinit,
                "{}".format("\n    ".join(met)),   
                "",
                "    #endregion",
                "",
                "#endregion"			
            ]   

            self.cls.append(baseclass)		
            self.cls.append("    #region Properties")   
            self.cls.append("".join(props))
            self.cls.append("    #endregion\n")            
            self.cls.append("\n    ".join(ctor))
            self.cls.append("")                        

    #endregion

    #region Methods    
    def isINT(self,Value):
        if isinstance(Value,int):
            return True
        else:
            return not re.fullmatch("[0-9]+", str(Value)) == None

    def isREAL(self,Value):        
        return not re.fullmatch("[0-9]+\.[0-9]+", str(Value)) == None

    def rowType(self,NewValue, OldType=None):	
        if OldType==None:
            if self.isREAL(NewValue):
                ret = "REAL"
            elif self.isINT(NewValue):
                if len(str(NewValue))> 5:
                    ret = "TEXT"
                else:
                    ret = "INT"		
            else:
                ret = "TEXT"
        else:
            if OldType in ["INT","REAL"] and not (self.isINT(NewValue) or self.isREAL(str(NewValue))):
                ret = "TEXT"
            elif OldType in ["INT"] and self.isREAL(str(NewValue)):
                ret = "REAL"		
            else:
                ret = OldType

        return ret

    def recurse(self ,data, name):
        if not isinstance(data, dict):
            pass      

        else:
            for k in data.keys(): 
                # print(k , data[k] )
                if isinstance(data[k], list):
                    if not name in list(self.mycls.keys()):
                        self.mycls[name] = {}
                        self.mylen[name] = {}

                    self.mycls[name][k] = []
                    self.mylen[name][k] = []

                    for l in data[k]:
                        self.recurse(l, k)
                
                elif hasattr(data[k], "keys"):
                    if not name in list(self.mycls.keys()):
                        self.mycls[name] = {}		
                        self.mylen[name] = {}
                    self.mycls[name][k] = "$"

                    if not k in list(self.mycls.keys()):
                        self.mycls[k] = {}		
                        self.mylen[k] = {}		

                    for i in data[k].keys():
                        if hasattr(data[k][i], "keys"):
                            self.mycls[k][i] = "$"
                            self.recurse(data[k][i], i)
                        else:    
                            self.mycls[k][i] = self.rowType(data[k][i])                        
                            self.mylen[k][i] = len(str(data[k][i]))

                else:
                    if not name in list(self.mycls.keys()):
                        self.mycls[name] = {}
                        self.mylen[name] = {}
                    
                    if not k in list(self.mycls[name]):
                        self.mycls[name][k] = self.rowType(data[k])
                        self.mylen[name][k] = len(str(data[k]))

                    else:
                        self.mycls[name][k] = self.rowType(data[k] , self.mycls[name][k])
                        if len(data[k]) > self.mylen[name][k]:
                            self.mylen[name][k] = len(str(data[k]))

    #endregion

#endregion


######################################################
## si@medatechuk.com
## 08/09/21
## https://github.com/MedatechUK/Medatech.APY
## 
## Usage:
## from apy import Request
## 
## request = Request()
## if request.method == "POST":
##    request.Response.data = request.data
## 
## if request.method == "GET":
##    request.Response.data = {"id": request.query("id","123")}
## 
## request.Response.Flush()

import os , sys, json
import importlib
import urllib.parse as urlparse
import xmltodict , dicttoxml
import pyodbc
from MedatechUK.oDataConfig import Config
import MedatechUK.apy
from MedatechUK.mLog import mLog
import inspect
from datetime import datetime

class Request:

    ## Ctor
    def __init__(self, **kwargs):
        
        ## Register for log
        self.log = mLog()
        
        ## Set Request defaults   
        self.content_type = "application/json"         
        self.content_length = 0                  
        self.ext = "json"
        self.data = {}                                                     
        self.config = {}
        self.serialtype = 'json'          

        try:            
            if kwargs.__contains__("method"):
                self.method = kwargs["method"]
            else:
                self.method = os.environ["REQUEST_METHOD"] 

            if kwargs.__contains__("environment"):
                self.environment = kwargs["environment"]
            else:
                self.environment = self.query("environment","")

            if kwargs.__contains__("endpoint"):
                self.endpoint = kwargs["endpoint"]
            else:
                self.endpoint = self.query("endpoint", "")   
            
            self.response = Response(request=self) 

        except Exception as e :
            ## Set the status/message of the response on error      
            self.log.logger.critical("Bad request.")
            self.log.logger.exception(e)
            self.Response.Status = 400
            self.Response.Message = "Internal Server Error"
            self.Response.data = {"error" : "Bad config: " + str(e)}

        if self.cont() : 
            try :  
                ## Generate the resonse object
                self.log.logger.debug("Handling {} /{}.".format( self.method , self.endpoint ))                
                
                ## Locate the root folder
                previous_frame = inspect.currentframe().f_back
                (filename, line_number, function_name, lines, index) = inspect.getframeinfo(previous_frame)
                self.path = os.path.dirname(filename)
                                    
                ## Split the endpoint into endpoint and extention
                #   where the endpoint contains a period
                if self.endpoint.find(".") > 0:                                      
                    self.ext = (self.endpoint.split(".")[-1]).lower()                 
                    self.endpoint = self.endpoint[0:len(self.endpoint)-(len(self.ext)+1)] 
                
                ## Generate the config object            
                self.config = Config(request=self)                         

            except Exception as e :
                ## Set the status/message of the response on error      
                self.log.logger.critical("Bad config.")
                self.log.logger.exception(e)
                self.Response.Status = 500
                self.Response.Message = "Internal Server Error"
                self.Response.data = {"error" : "Bad config: " + str(e)}

        if self.environment != "" and self.cont() :   
            ## Is it a valid environment?
            if not self.config.CheckEnviroment():
                self.log.logger.critical("Invalid environment {}".format(self.config.environment))
                self.Response.Status = 500
                self.Response.Message = "Internal Server Error"
                self.Response.data = {"error" : "Invalid environment [{}]".format(self.config.environment)} 

        ## Set the Content-Type of the request
        #   for GET            
        if self.method == "GET" and self.cont() :                                
            #   Set the response content type based on the file extention    
            #   .xml / .ashx
            if self.ext == "xml" or self.ext == "ashx":
                self.content_type = "application/xml"

            #   .json / unknown enpoint file extention
            else:
                # Default
                self.content_type = "application/json"  
            self.log.logger.debug("GET Content type: [{}].".format( self.content_type ))

            ##  Import a script to handler the request
            #   Check file exists
            if os.path.isfile(self.endpoint + ".py"):
                self.Inject()
            
            ## Load from database
            elif self.environment !="" :
                try:
                    self.log.logger.debug("Opening DB query: [{}] in [{}].".format( self.endpoint , self.environment ))
                    cnxn = self.config.cnxn()
                    crsr = cnxn.cursor() 

                    crsr.execute(
                        "select SO.OBJECT_ID as [ObjectID], " +
                        "SCHEMA_NAME(SCHEMA_ID) + '.' + SO.name AS [ObjectName] " +
                        "From sys.objects AS SO " +
                        "INNER JOIN sys.parameters AS P " +
                        "On SO.OBJECT_ID = P.OBJECT_ID " +
                        "WHERE 0=0 " +
                        "And SO.TYPE IN ('FN') " +
                        "And (TYPE_NAME(P.user_type_id)='xml') " +
                        "And (LOWER(SO.name)=LOWER('"+ self.endpoint +"')) " +
                        "And P.is_output=1 "
                    )

                    row = crsr.fetchone()                   
                    sql = "SELECT " + row.ObjectName + " ("
                    crsr.execute(            
                        "SELECT	" +
                        "	P.name AS [ParameterName],	" +
                        "	TYPE_NAME(P.user_type_id) AS [ParameterDataType] " +
                        "FROM sys.objects AS SO	" +
                        "	INNER JOIN sys.parameters AS P 	" +
                        "	ON SO.OBJECT_ID = P.OBJECT_ID	" +
                        "WHERE 0=0	" +
                        "	And SO.OBJECT_ID = "+ str(row.ObjectID) +
                        "	And P.is_output=0" +
                        "order by parameter_id"
                    )
                                                        
                    for row in crsr.fetchall() :
                        if row.ParameterDataType in ["char", "varchar", "text", "nchar", "nvarchar", "ntext"]:
                            sql += "'" + self.query(row.ParameterName[1:],"") + "', "
                        else:
                            sql += self.query(row.ParameterName[1:],"0") + ", "
                    
                    sql = sql[0:len(sql)-2]                    
                    crsr.execute(sql + ')')
                    row = crsr.fetchone()
                    self.Response.data = xmltodict.parse(row[0])                
                
                except Exception as e:
                    self.log.logger.critical("Database error.")
                    self.log.logger.exception(e)
                    self.Response.Status = 404
                    self.Response.Message = "Not found."
                    self.Response.data = {"error" : self.endpoint + '.' + self.ext + " not found.", "dberror" : str(e)}                       
            else:
                self.log.logger.critical("Handler not found: [{}].".format( self.endpoint + ".py" ))
                self.Response.Status = 404
                self.Response.Message = "Not found."
                self.Response.data = {"error" : self.endpoint + '.' + self.ext + " not found.", "dberror" : str(e)} 
                
        ##  Set the Content-Type of the request
        #   for POST            
        elif self.method == "POST" and self.cont():
            if not os.path.isfile(self.endpoint + ".py"):
                self.log.logger.critical("handler not found.")
                self.content_type = "application/json"     
                self.Response.Status = 404
                self.Response.Message = "handler not found."
                self.Response.data = {"error" : "handler not found"} 

            if self.cont():
                #   Set the response content type based on request content type
                if kwargs.__contains__("data"):
                    if kwargs.__contains__("content_type"):
                         self.content_type = kwargs["content_type"]
                    data = json.dumps(kwargs["data"])
                else:
                    try:
                        self.content_Length = int(os.environ.get('CONTENT_LENGTH', '0'))      
                    except:
                        self.content_Length = 0

                    if self.content_Length > 0 :                        
                        self.content_type = os.environ['HTTP_CONTENT_TYPE']  

                        cl = self.content_Length
                        data = "" # sys.stdin.read(content_Length)   ## Easier, didn't work                 
                        while cl > 0:
                            o = sys.stdin.read(1)       
                            data += o
                            cl += -1                        
                            if "\\n" in ascii(o):    
                                ## Content length returns 2 chars (cr lf) for \n                      
                                #  BUT stdin reads BOTH characters as a single char
                                #  causing a buffer overrun.
                                #  This removes the extra characters.
                                cl += -1  

                    else:
                        self.log.logger.critical("Bad request: [Missing Content].")
                        self.content_type = "application/json"     
                        self.Response.Status = 400
                        self.Response.Message = "Bad Request"
                        self.Response.data = {"error" : "No data in request"}   

                    #   Check for valid content type
                    if self.content_type != "application/xml" and self.content_type != "application/json" :
                        self.log.logger.critical("Bad request: [Invalid Content type].")
                        self.content_type = "application/json"     
                        self.Response.Status = 400
                        self.Response.Message = "Bad Request"
                        self.Response.data = {"error" : "Invalid Content type. Use application/xml or application/json"}            
                                    
                ##  Deserialise to self.data if no previous error
                if self.cont() :
                    self.log.logger.debug("Deserialising Content type: [{}].".format( self.content_type ))
                    try:                                          
                        if self.content_type=="application/json" :          
                            self.data = json.loads(data)
                            self.serialtype = 'json'

                        if self.content_type=="application/xml" : 
                            t = xmltodict.parse(data)
                            self.data = json.loads(json.dumps(t[list(t)[0]]))
                            self.serialtype = 'xml'

                        self.log.logger.info("Received [{}]: {}".format(self.serialtype , json.dumps(self.data, sort_keys=False, indent=4)))

                    except Exception as e:                    
                        # Invalid data
                        self.log.logger.critical("Bad request: [Invalid data].".format( str(e) ))
                        self.Response.Status = 400
                        self.Response.Message = "Invalid POST"
                        self.Response.data = {"error" : str(e)}
                    
                ##  Inject a handler for the request
                #   if the handler exists.
                if self.cont():
                    self.Inject()

    def Inject(self):
        handler = {}
        try:
            # Import the handler
            self.log.logger.debug("Injecting handler: [{}].".format( self.endpoint + ".py" ))            
            handler = importlib.import_module(self.endpoint)

        except Exception as e :
            self.log.logger.critical("Injection Failure [{}]: {}.".format( self.endpoint + '.' + self.ext , str(e) ))
            self.Response.Status = 500
            self.Response.Message = "Injection Failure"
            self.Response.data = {"error" : str(e), "handler": self.endpoint}

        try:
            # Process the request with the loaded handler
            self.log.logger.debug("Running ProcessRequest method in: [{}].".format( self.endpoint + ".py" ))   
            handler.ProcessRequest(self)

        except Exception as e :      
            self.log.logger.critical("Injection Failure [{}]: {}.".format( self.endpoint + '.' + self.ext , str(e) ))                  
            self.Response.Status = 500
            self.Response.Message = "Handler error"
            self.Response.data = {"handler": self.endpoint + '.' + self.ext , "error" : str(e)}

    # Get params from the query String
    def query(self, name , default):
        ret = ''
        for k in urlparse.parse_qs(os.environ["QUERY_STRING"]):                        
            if str(k).lower() == str(name).lower():                
                for i in range(len(urlparse.parse_qs(os.environ["QUERY_STRING"])[k])):
                    if len(ret) > 0:
                        ret += ","
                    ret += urlparse.parse_qs(os.environ["QUERY_STRING"])[k][i-1]
                    
        if len(ret) > 0:
            self.log.logger.debug("URL Query: ?{}={}".format(name , ret))
            return ret
        else:
            self.log.logger.warn("URL Query: ?{}=null. default={}".format(name , default))
            return default  

    # Returns true is the response status is 2**
    def cont(self):
        ret = (self.response.Status >= 200 and self.response.Status <= 299)
        #if ret:
        #    self.log.logger.debug("[{}] Continue.".format( self.Response.Status ))        
        return ret

class Response:

    ## Ctor
    def __init__(self, **kwargs):
        
        ## Register for log
        self.log = mLog()   

        if kwargs.__contains__("request"):
            self.request = kwargs["request"]
            self.request.response = self

        self.Status = 200         
        self.Message = "OK"                          
        self.data = {}
    
    ## Flush method: Send response to the client
    def Flush(self):        

        self.log.logger.debug("Flushing response.")

        ## redirecting?
        if self.Status == 302:
            self.log.logger.debug("Redirecting response to [{}].".format( self.Message ))
            print("HTTP/1.1 {} Found".format(str(self.Status)))
            print("Location: {}".format(self.Message))
            print("")

        else:
            ## Write Headers        
            self.ContentHeader()
            self.log.logger.debug("Write [{}] response.".format( self.request.content_type ))             

            ## Write self.data to the response
            if self.request.content_type=="application/xml" :
                ## In XML                   
                print(dicttoxml.dicttoxml(self.data).decode('utf-8'))

            else :
                ## In JSON
                print(json.dumps(self.data, indent=4))

    def redirect(self, url):
        ## Set redirect
        self.Status = 302
        self.Message = url

    ## Output internals as response headers for debugging
    def ContentHeader(self):

        ## Return the error code and message
        #  https://www.w3.org/Protocols/rfc2616/rfc2616-sec6.html
        print('HTTP/1.1 {} {}'.format(str(self.Status), self.Message))
        print("Content-Type: {}".format(self.request.content_type))

        try :
            ## Content header        
            print("Environment: {}".format(self.request.environment))
            print("Endpoint: {}".format(self.request.endpoint))
            print("Endpoint-Type.:{}".format(self.request.ext)  )          
            print("oDataHost: {}".format(self.request.config.oDataHost.split("//")[-1]))
            print("tabulaini: {}".format(self.request.config.tabulaini))
            print("db: {}".format(self.request.config.connstr))

        finally:
            print('')   



######################################################
## si@medatechuk.com
## 14/08/21
## https://github.com/MedatechUK/Medatech.APY
##
## Create a new load object
## request = POST   ) Either
## f = file         ) Or
## o = object       ) Or
## env = company    MANDATORY is file or Object
## c = config       MANDATORY: created if missing 
## ltype = LoadType MANDATORY: the type of loading
## ex = exclude     Exlude namespaces from parse
## 
## Usage:
## l = Load(
##     ltype = "TST",
##     f = sys.argv[1], ## First param on command line
##     c = sys.argv[2] ## Second param on command line    
## )  
## l.save('odata.json') ## Save to file
## l.post() ## Post to Priority
##
## Constants.py:
##  oDataHost ="walrus.ntsa.uk"
##  tabulaini ="tabula.ini"
##  ouser ="apiuser"
##  opass ="123456"
##  Environment = "wlnd"

from datetime import datetime
import sys
import json
from types import SimpleNamespace
import os
import os.path
import uuid
from http.client import HTTPSConnection
from base64 import b64encode
import inspect

from MedatechUK.mLog import mLog
from MedatechUK.oDataConfig import Config
import MedatechUK.PriDate

class oRecord:
    ## Ctor
    def __init__(self , recordtype , line):

        ## Set Default Property values
        self.recordtype = recordtype
        self.line = line
        ## An arraty of Name/Value pairs for Properties in this record
        self.oProps = [] 

class oProp:
    ## Ctor
    def __init__(self , name , value):
        
        ## Set Default Property values
        self.name = name
        self.value = value

class Rec:    
    ## Ctor
    def __init__(self , rt, name):
        
        ## Set Default Property values
        self.rt = rt ## The record type
        self.name = name ## The record name
        self.props = [] ## The array or propeerties in this Record Type

        # Init NEXT values for each Priority type
        self.text = 0
        self.real = 0
        self.int = 0
        self.bool = 0
    
    ## Methods called by child props for NEXT type
    def nexttext(self):
        self.text = self.text + 1
        return str(self.text)
    def nextreal(self):
        self.real = self.real + 1
        return str(self.real)
    def nextint(self):
        self.int = self.int + 1
        return str(self.int)
    def nextbool(self):
        self.bool = self.bool + 1
        return str(self.bool)

class Prop:   

    ## Ctor
    def __init__(self , property , type , parent):
        
        ## Set Default Property values
        self.property = property
        self.type = 'unknown'
        
        ## Use existing type if parent record already 
        ## contains this property.
        if not len(parent.props) == 0:
            for n in parent.props:
                if n.property == property:
                    self.type = n.type
                    return ## quit __init__

        ## It's a new property in the the record.
        ## Call the records NEXT function for the 
        ## next type number.
        if type == float:
            self.type = "REAL" + parent.nextreal()
        elif type == str:            
            self.type = "TEXT" + parent.nexttext()
        elif type == int:            
            self.type = "INT" + parent.nextint()
        elif type == datetime:            
            self.type = "INT" + parent.nextint()            
        elif type == bool:
            self.type = "CHAR" + parent.nextbool()           
        else:
            print("Warning! Unknown type: " + type)            

class Load:    

    ## Ctor
    def __init__(self, **kwargs): 
        
        self.log = mLog()    

        ## Set Default Property values       
        self.line = 1
        self.records = [] ## Holds Record type data        
        self.odata = [] ## Holds the oData records
        self.config = {} ## Hold the configuration
        self.data = {} ## Hold the oData Commands
        self.patch = {} ## Hold the oData PATCH Commands
        self.BubbleID = str(uuid.uuid4()) ## The transaction identifier
        self.json = {} # The json data from file/object
        self.ex = {} # The excluded namespaces / columns
        self.path = ""

        ## Determine data source
        for arg in kwargs.keys():

            ## Data from request
            if arg.lower() == 'request' :   
                self.log.logger.debug("Opening [{}] Request for [{}] ...".format(kwargs[arg].content_type , kwargs[arg].endpoint))                
                try:                                                                   
                    self.path = kwargs[arg].path
                    self.json = json.loads(json.dumps(kwargs[arg].data), object_hook=lambda d: SimpleNamespace(**d))                                                     
                    self.oDataHost = kwargs[arg].config.oDataHost
                    self.url = '/odata/priority/{}/{}/ZODA_TRANS'.format(kwargs[arg].config.tabulaini , kwargs[arg].config.environment)
                    self.headers = { 
                        'Authorization' : 'Basic %s' %  b64encode(bytearray(kwargs[arg].config.ouser + ":" + kwargs[arg].config.opass,'ascii')).decode("ascii") ,
                        'Content-Type': 'application/json',
                        "User-Agent": "MedatechUK Python Client",
                    }

                except Exception as e :     
                    self.log.logger.exception(e)                   
                    kwargs[arg].Response.Status = 500
                    kwargs[arg].Response.Message = "Bad Config: " + str(e)

            ## The environment is specified
            if arg.lower() == 'env' :           
                self.log.logger.debug("Setting environment [{}]...".format(kwargs[arg] )) 

                ## Locate the root folder
                previous_frame = inspect.currentframe().f_back
                (filename, line_number, function_name, lines, index) = inspect.getframeinfo(previous_frame)
                self.path = os.path.dirname(filename)           

                c = Config(env=kwargs[arg] , path=self.path)               
                self.oDataHost = c.oDataHost
                self.url = '/odata/priority/{}/{}/ZODA_TRANS'.format(c.tabulaini , c.environment)
                self.headers = { 
                    'Authorization' : 'Basic %s' %  b64encode(bytearray(c.ouser + ":" + c.opass,'ascii')).decode("ascii") ,
                    'Content-Type': 'application/json',
                    "User-Agent": "MedatechUK Python Client",
                }  
        
        self.log.logger.debug("Running in [{}]...".format( self.path )) 
        for arg in kwargs.keys():

            ## The excluded namespaces / columns
            if arg.lower() =='ex' :
                self.ex = kwargs[arg]      

            ## Data from file
            if arg.lower() == 'f' :                 
                try:
                    self.log.logger.debug("Opening from file: [{}] ... ".format(kwargs[arg]))
                    with open(self.path + "\\" + kwargs[arg], "r") as o:   
                        self.json = json.loads(o.read(), object_hook=lambda d: SimpleNamespace(**d)) 

                except Exception as e :       
                    self.log.logger.exception(e)                 
                    kwargs[arg].Response.Status = 500
                    kwargs[arg].Response.Message = "Bad Config: " + str(e)

            ## Data from object
            if arg.lower() == 'o' :  
                try:               
                    self.log.logger.debug("Opening from object ... ")
                    self.json = json.loads(kwargs[arg].read(), object_hook=lambda d: SimpleNamespace(**d))    

                except Exception as e : 
                    self.log.logger.exception(e)
                    kwargs[arg].Response.Status = 500
                    kwargs[arg].Response.Message = "Bad Config: " + str(e)

        ## Recurse through data to create config
        self.log.logger.debug("Make loading config.")
        self.makeConfig(self.json, 'root')           
        
        ## Write config (if missing)
        self.writeConfig(kwargs['c'])      
                
        ## Load the config
        with open(self.path + "\\" + kwargs['c'], "r") as f:            
            self.config = json.loads(f.read(), object_hook=lambda d: SimpleNamespace(**d))   
                
        ## print("Parsing object ... ")  
        self.log.logger.debug("Parsing loading.")      
        self.parse(self.json, 'root')
        
        ## Build POST request body
        self.data['TYPENAME'] = kwargs['ltype']
        self.data['BUBBLEID'] = self.BubbleID
        self.data['ZODA_LOAD_SUBFORM'] = []
        for r in self.odata:        
            self.data['ZODA_LOAD_SUBFORM'].append({})
            self.data['ZODA_LOAD_SUBFORM'][-1]['RECORDTYPE'] = str(r.recordtype)
            for p in r.oProps:
                self.data['ZODA_LOAD_SUBFORM'][-1][p.name] = p.value  
        
        ## Build PATCH request body  
        self.patch['COMPLETE'] = "Y"

    ##                
    ## Add a path to the loading
    def Add(self , path):
        thisrt = 0
        for r in self.records:

            if r.rt > thisrt:
                ## Count RTs
                thisrt = r.rt
            
            if r.name == path:
                ## Found record type, return
                return r

        ## Record Type not found: Add.
        self.records.append( Rec( thisrt + 1 , path) )
        return self.records[-1]

    ## Return the RecordType of a path
    def rt(self, path):
        for r in self.records: 
            if r.name == path:
                return r.rt
        return -1

    ## Return the property
    def prop(self, rt, name, v):
        for property, value in vars(self.config.config.row).items():            
            if int(property) == int(rt):                
                for source, dest in vars(value).items():
                    if source == name:
                        return oProp(dest , v)

    ## Recurse json to create config data
    def makeConfig(self, x, path):            
        rt = []    
        try:
            for property, value in vars(x).items():  
                # Excluded?
                if not property in self.ex:             
                    # If it's not a namespace or list      
                    if type(value) is not SimpleNamespace and type(value) is not list:
                        # Add the column at this location
                        if MedatechUK.PriDate.isDate(value):
                            # If the value is a date then initialise with data value
                            self.Add(path).props.append(Prop(property , type(datetime.now()), self.Add(path)))                
                        else:
                            self.Add(path).props.append(Prop(property , type(value), self.Add(path)))                    

            for property, value in vars(x).items(): 
                # Excluded?
                if not property in self.ex:            
                    th = path + "." + property           
                    # IS it a namespace? 
                    if type(value) is SimpleNamespace:              
                        if th not in rt:
                            rt.append(th)            
                        self.makeConfig(value, th)
                    
            for property, value in vars(x).items():       
                # Excluded?
                if not property in self.ex:
                    th = path + "." + property  
                    # IS it a List?
                    if type(value) is list:
                        for i in value:             
                            if th not in rt:
                                rt.append(th)                                           
                            self.makeConfig(i, th) 
        
        except Exception as e:            
            self.Add(path).props.append(Prop("TEXT1" , type("TEXT1"), self.Add(path)))   

    ## Write the config file
    def writeConfig(self , fn):
    
        ## Does the config file already exist?
        if not os.path.isfile(fn):
    
            ## Create the config file
            ## print("Creating config [{}] ... ".format(fn))
            rw = {}
            rw['config'] = {}
            rw['config']['rt'] = {}
            rw['config']['row'] = {}    
            for r in self.records:
                rw['config']['rt'][r.rt] = r.name
                rw['config']['row'][r.rt] = {}
                for p in r.props:
                    rw['config']['row'][r.rt][p.property] = p.type
                
                with open(fn, 'w') as f:
                    json.dump(rw, f)                     

    def configRows(self , path):
        for property, value in vars(self.config).items():              
            for property, value in vars(value).items():
                if property == 'row':
                    for ROWID, value in vars(value).items():
                        if int(ROWID) == int(self.rt(path)):
                            return value

    ## Parse the data using the config
    def parse(self , x , path):

        rt = []
        addline = False

        for p, v in vars(x).items():    
            ## Not namespace or list - a column    
            if type(v) is not SimpleNamespace and type(v) is not list and not p in self.ex:
                if not addline:
                    ## Create new oRecord in the oData, if one doesn't exist
                    addline = True
                    self.odata.append(oRecord(self.rt(path) , self.line))            

                for property, value in vars(self.configRows(path)).items():
                    if property == p:                   
                        if value[0:3] == "INT" and MedatechUK.PriDate.isDate(v):                            
                            self.odata[-1].oProps.append(oProp(value, MedatechUK.PriDate.IntDate(v)))
                        elif value[0:4] == "CHAR":
                            if v:
                                self.odata[-1].oProps.append(oProp(value,"Y"))
                            else:
                                self.odata[-1].oProps.append(oProp(value,"N"))
                        else:
                            self.odata[-1].oProps.append(oProp(value,v))
    
        if not p in self.ex:
            ## Increment line counter if it's a new line
            if addline:
                self.line += 1
                
        for property, value in vars(x).items():   
            # Excluded?
            if not property in self.ex:                       
                th = path + "." + property    
                # Am I a namespace?        
                if type(value) is SimpleNamespace:              
                    if th not in rt:
                        rt.append(th)            
                    self.parse(value, th)
                
        for property, value in vars(x).items():      
            # Excluded?
            if not property in self.ex:                 
                th = path + "." + property  
                # Am I a list?        
                if type(value) is list:
                    for i in value:             
                        if th not in rt:
                            rt.append(th)   
                        try:                 
                            test = vars(x).items()
                            self.parse(i, th)
                        except:       
                            self.odata.append(oRecord(self.rt(th) , self.line))                             
                            self.odata[-1].oProps.append(oProp("TEXT1", i ))        
                            self.line += 1                            
            
    ## Save the odata to a file
    def save(self , fn):
        self.log.logger.debug("Saving to [{}] ... ".format(fn))
        with open(fn, 'w') as f:
            json.dump(self.data, f)            

    ## Post to Priority
    def post(self, Response):
        self.log.logger.debug("POSTing to [{}{}] ".format( self.oDataHost , self.url ))                 
        self.log.logger.debug("Headers:\n{}".format( json.dumps(self.headers ,  indent = 4) ))
        self.log.logger.debug("Data:\n{}".format( json.dumps(self.data, indent = 4) ))
        
        c = HTTPSConnection(self.oDataHost)  
        c.request( 
            'POST', 
            self.url , 
            headers=self.headers, 
            body=json.dumps(self.data) 
        )
        res = c.getresponse()            
        if res.status == 201: # Created
            self.log.logger.debug("[{}] OK".format( res.status ))
            data = json.loads(res.read())             

            c.request( 
                'PATCH', 
                self.url + "(BUBBLEID='"+ data['BUBBLEID'] + "',LOADTYPE=" + str(data['LOADTYPE']) + ")", 
                headers=self.headers, 
                body=json.dumps(self.patch) 
            )
            
            self.log.logger.debug("PATCHing to [{}] ... ".format( self.url + "(BUBBLEID='"+ data['BUBBLEID'] + "',LOADTYPE=" + str(data['LOADTYPE']) + ")" ))  
            res = c.getresponse()

            if res.status != 200: # PATCHed
                self.log.logger.critical("[{}] Fail: {}".format( res.status , res.reason ))
                Response.Status = res.status   
                Response.Message = "PATCH Failed: " + res.reason  
                # If the response is text, create a response with the text
                if res.getheader("Content-Type","").find("text/plain") > -1:                             
                    er = str(res.read().decode('utf-8'))
                    Response.data = {"error": er }     
                    self.log.logger.critical("{}".format( er ))

                else: # Create reponse from json 
                    Response.data = json.load(res)

            else:
                ## Sucsess!
                self.log.logger.debug("[{}] {}".format( res.status , res.reason ))
                Response.Status = res.status
                Response.Message = res.reason
                Response.data = json.load(res)

        else:   
            Response.Status = res.status
            Response.Message = "POST Failed: " + res.reason   
            self.log.logger.critical("[{}] Fail: {}".format( res.status , res.reason ))

            # If the response is text, create a response with the text         
            if res.getheader("Content-Type","").find("text/plain") > -1:                             
                er = str(res.read().decode('utf-8'))
                Response.data = {"error": er }     
                self.log.logger.critical("{}".format( er ))

            else: # Create reponse from json 
                Response.data = json.load(res)   

## Response structure used for non request calls
class oResponse:

    ## Ctor
    def __init__(self):
                  
        self.Status = 200         
        self.Message = "OK"                          
        self.data = {}


from datetime import datetime
from dateutil.parser import parse

def isDate(str):
    try:
        parse(str)
        return True
    except:
        return False

def IntDate(str):    
    d = parse(str)    
    return int(
        (datetime(
            d.year, 
            d.month, 
            d.day, 
            d.hour, 
            d.minute) 
        - datetime(1988, 1, 1)).total_seconds() / 60
    )
    



import sys
from MedatechUK.mLog import mLog
import os
from os import listdir
from os.path import exists, isfile, join
import sys
import subprocess
from os.path import exists

#region Command Line Arguments            
class clArg():

    def __init__(self, **kwargs):
        log = mLog()
        self._kw = {}
        self._a = []
        last = None
        if kwargs.__contains__("args"):
            sys.argv = kwargs["args"]            
        for arg in sys.argv[:0:-1] :
            if arg[0] in {'-', '/'} :                
                while arg[0] in {'-', '/'} :
                    arg = arg[1:]
                self._kw[arg] = last
                last = None
            else:
                if last != None:
                    self._a.append(last)
                last = arg
        if last != None:
            self._a.append(last)

        for k in self._kw:
            log.logger.debug("Key: [{}] = {}".format(k.lower() , self._kw[k]))
        for a in self._a:
            log.logger.debug("Arg [{}]".format( a ) )

    def kwargs(self):        
        return self._kw
    
    def args(self):
        return self._a[::-1]
    
    def argExists(self, id):
        return exists(self._a[::-1][id])
    
    def byName(self, key):
        for k in self._kw:
            for i in key:
                if k.upper() == i.upper():
                    return self._kw[k]
        return None

#endregion

#region folder watcher
class folderWatch():
    
    def __init__(self , **kwargs):
        
        self.log = mLog()
        
        #region cl validation
        if not kwargs.__contains__("folder"):
            self.log.logger.critical("Folder not specified.")
            raise NameError("Folder not specified.")
        if not kwargs.__contains__("handler"):
            self.log.logger.critical("Handler not specified.")
            raise NameError("Handler not specified.")            
        if not kwargs.__contains__("env"):
            self.log.logger.critical("Environment not specified.")
            raise NameError("Environment not specified.")                        
        if not exists(kwargs['folder']):
            self.log.logger.critical("Folder {} not found.".format(kwargs['folder']))
            raise NameError("Folder {} not found.".format(kwargs['folder']))
        if not exists(kwargs['handler']):
            self.log.logger.critical("handler [{}] not found.".format(kwargs['handler']))
            raise NameError("handler [{}] not found.".format(kwargs['handler']))

        #endregion

        #region Set local properties from commandline
        self._folder = kwargs['folder'].rstrip('\\') + "\\"
        self._handler = kwargs['handler']
        self._env = kwargs['env']        
        
        if not kwargs.__contains__('ext'):
            self._ext = None
            self.log.logger.info("fWatch folder {} with handler {} in env {}.".format(self._folder  , self._handler , self._env ))
        else :            
            self._ext = kwargs['ext']        
            self.log.logger.info("fWatch folder {} for {} with handler {} in env {}.".format(self._folder  ,self._ext, self._handler , self._env ))

        #endregion

    def EnvStr(self):
        return "-e {}".format(self._env)

    def filePath(self,f):
        return "{}{}".format(self._folder , f)

    def CMD(self, cwd, f):
        return ( 
            "{} {} {} {}".format (
                self._handler ,
                self.EnvStr() ,
                "-cwd {}".format(cwd) ,
                self.filePath(f)        
            )
        )

    def isFileLocked(self, filePath):
        '''
        Checks to see if a file is locked. Performs three checks
            1. Checks if the file even exists
            2. Attempts to open the file for reading. This will determine if the file has a write lock.
                Write locks occur when the file is being edited or copied to, e.g. a file copy destination
            3. Attempts to rename the file. If this fails the file is open by some other process for reading. The 
                file can be read, but not written to or deleted.
        @param filePath:
        '''
        if not (os.path.exists(filePath)):
            return False
        try:
            f = open(filePath, 'r')
            f.close()
        except IOError:
            return True

        lockFile = filePath + ".lckchk"
        if (os.path.exists(lockFile)):
            os.remove(lockFile)
        try:
            os.rename(filePath, lockFile)
            # sleep(1)
            os.rename(lockFile, filePath)
            return False
        except WindowsError:
            return True

    def files(self):
        onlyfiles = [f for f in listdir(self._folder) if isfile(join(self._folder, f))
            and not self.isFileLocked(self._folder + '\\' + f)]
        if self._ext == None :
            return onlyfiles
        else :
            return [f for f in onlyfiles 
                if os.path.splitext(f)[1].lstrip('.').upper()==self._ext.lstrip('.').upper()
                    and not self.isFileLocked(self._folder + '\\' + f)]
    
    def check(self, cwd):
        for f in self.files():  
            self.log.logger.info("Found file {}.".format(self.filePath(f))) 
            self.log.logger.info(               
                "shell: {}".format(self.CMD(cwd , f))
            )         
            subprocess.call(
                self.CMD(cwd , f)
                , shell=False
            )
            self.log.logFile(self.filePath(f))

#endregion

#region Examples
if __name__ == '__main__' :

    l = mLog()
    l.start( "C:\pyedi", "DEBUG" )
    l.logger.debug("Starting {}".format(__file__)) 

    ## Get command line arguments
    arg = clArg()
    #print(arg.kwargs())
    #print(arg.args()[0])
    #print(arg.byName(['arg1']))

    ## Watch a folder
    fs = folderWatch(
        folder="M:\\python\\apy\\SolidWorks\\" , 
        handler="M:\\python\\apy\\solidworks.exe" , 
        env="wlnd" , 
        ext="xml"
    )
    fs.check("C:\\pyedi")

#endregion    


import os , time , sys
from pathlib import Path

import win32serviceutil , win32service , win32event, servicemanager
import configparser , debugpy

from MedatechUK.mLog import mLog
from MedatechUK.cl import clArg
from MedatechUK.oDataConfig import Config

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

class AppSvc (win32serviceutil.ServiceFramework):

#region "Properties"
    @property
    def Main(self):
        try: return self._Main
        except: return None
    @Main.setter
    def Main(self, value):
        self._Main = value

    @property
    def Init(self):
        try: return self._Init
        except: return None
    @Init.setter
    def Init(self, value):
        self._Init = value

    @property
    def Folder(self):
        try: return self._Folder
        except: return os.getcwd()
    @Folder.setter
    def Folder(self, value):
        if os.path.isdir(str(value)):
            self._Folder = value
        else: raise "Invalid folder spec."

#endregion

#region "Ctor"
    def __init__(self,args):

        win32serviceutil.ServiceFramework.__init__(self,args)
        self.hWaitStop = win32event.CreateEvent(None,0,0,None)
        self.stop = False

        self.settingsini = str(self.Folder) + "\\settings.ini"
        os.chdir(self.Folder)

        #region "Verify .ini"
        if os.path.isfile(self.settingsini):
            config = configparser.ConfigParser()
            config.read(self.settingsini)
            save = False
            if 'debug' not in config.sections():
                config['debug'] = {}
                save = True
            if not config['debug'].__contains__("verbosity"):
                config['debug']['VERBOSITY'] = 'DEBUG'
                save = True
            if not config['debug'].__contains__("port"):
                config['debug']['PORT'] = '5678'
                save = True
            if not config['debug'].__contains__("FORCE"):
                config['debug']['FORCE'] = 'ON'
                save = True                
        else:
            save = True
            config = configparser.ConfigParser()
            config['debug'] = {}
            config['debug']['VERBOSITY'] = 'DEBUG'
            config['debug']['PORT'] = '5678'
            config['debug']['FORCE'] = 'ON'
        
        if save :
            with open(self.settingsini, 'w') as configfile:
                config.write(configfile)

        #endregion

        config = configparser.ConfigParser(dict_type=AttrDict)
        config.read(self.settingsini)
        self.config = config._sections

        self.log = mLog()
        self.log.start( str(self.Folder) , self.config.debug.verbosity )
        self.log.logger.info("Using Settings file: [{}]".format( self.settingsini ))

        self.clArg = clArg(args=args)
        self.debug = ( 'debug' in self.clArg.kwargs() or self.config.debug.force.lower() == "on" )
        if self.debug:
            self.log.logger.info("debugpy listening on: [{}]".format(self.config.debug.port))
            debugpy.configure(python=str(Path(sys.executable).parent) + "\\python.exe")
            debugpy.listen( int( self.config.debug.port ) )
            # debugpy.wait_for_client() 
            # debugpy.breakpoint() 

        self.debuginit = False
        if self.Init != None:
            if self.debug :
                if 'debug'.lower() in self.clArg.kwargs() :
                    self.debuginit = self.clArg.kwargs()["debug"] == "init".lower()
                if self.debuginit:
                    debugpy.wait_for_client()                    
            self.Init(self)

        try:
            Config(
                env = 'test'
                , path = str(self.Folder)
            )
        except Exception as error:
            self.log.logger.critical(error)
            self.stop = True
            return

        self.oDataConfig = Config(
                env = self.config.odata.env
                , path = str(self.Folder)
            )
        
        if self.Main == None:
            self.log.logger.critical("No Main() entry point.")
            self.stop = True
            return
                
        self.log.logger.info("Intialising Service [{}] with Verbosity [{}]...".format(
            self._svc_name_
            , self.config.debug.verbosity
        ))

        self.log.logger.info("Starting Service [{}]...".format(self._svc_name_ ))
        servicemanager.LogMsg(servicemanager.EVENTLOG_INFORMATION_TYPE,
                        servicemanager.PYS_SERVICE_STARTED,
                        (self._svc_name_,''))
#endregion

#region "Methods"
    def SvcStop(self):
        self.log.logger.info("Service [{}] Stopping...".format(self._svc_name_ ))
        if self.debug: debugpy.wait_for_client.cancel()

        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)
        self.stop = True

    def SvcDoRun(self):
        while not self.stop:
            try:
                if self.debug: debugpy.wait_for_client()
                self.Main(self)

            except Exception as error:
                self.log.logger.critical(error)
                raise

            for i in range(50): # 5 seconds
                if not self.stop:
                    time.sleep(.1)
                if self.stop: break

#endregion


from MedatechUK.Serial import SerialBase , SerialT , SerialF

class xmlTransactions(SerialBase) :
    
    #region Properties
    @property
    def transactions(self): 
        return self._transactions
    @transactions.setter
    def transactions(self, value):
        self._transactions = []
        for i in range(len(value)):
            if len(value) > 1 :
                self._transactions.append(xmlTransactionsTransaction(**value["transaction"][i]))
            else :
                self._transactions.append(xmlTransactionsTransaction(**value["transaction"]))
    
    #endregion

    #region "ctor"
    def __init__(self,  **kwargs): 

        #region "Property defaults"
        self.transactions = []
        
        #endregion

        #region "Set Meta info"
        SerialBase.__init__(self , SerialF(fname="ZODA_TRANS", rt=1, typename="ORD"), **kwargs)          
        
        SerialT(self, "rt")
        SerialT(self, "bubbleid")
        SerialT(self, "typename")

        #endregion
    
    #endregion

class xmlTransactionsTransaction(SerialBase) :
    
    #region Properties
    @property
    def document(self): 
        return self._document
    @document.setter
    def document(self, value):    
        self._document = {}
        self._document = xmlTransactionsTransactionDocument(**value)
    
    @property
    def date(self): 
        return self._date
    @date.setter
    def date(self, value):
        self._date = value

    @property
    def type(self): 
        return self._type
    @type.setter
    def type(self, value):
        self._type = value
    
    @property
    def vaultname(self): 
        return self._vaultname
    @vaultname.setter
    def vaultname(self, value):
        self._vaultname = value
    
    #endregion

    #region "ctor"
    def __init__(self,  **kwargs): 

        #region "Property defaults"  
        self._document = {}
        self._date = 0
        self._type = ""
        self._vaultname = ""
        
        #endregion

        #region "Set Meta info"
        SerialBase.__init__(self , SerialF(fname="Transaction", rt=2) , **kwargs)           
        
        SerialT(self, "rt")
        SerialT(self, "date" , pCol="INT1" , pType="INT")
        SerialT(self, "type" , pCol="TEXT2")
        SerialT(self, "vaultname", pCol="TEXT3")

        #endregion
    
    #endregion

class xmlTransactionsTransactionDocument(SerialBase) :
    
    #region Properties
    @property
    def configuration(self): 
        return self._configuration
    @configuration.setter
    def configuration(self, value):        
        self._configuration = xmlTransactionsTransactionDocumentConfiguration(**value)
    
    @property
    def aliasset(self): 
        return self._aliasset
    @aliasset.setter
    def aliasset(self, value):
        self._aliasset = value

    @property
    def id(self): 
        return self._id
    @id.setter
    def id(self, value):
        self._id = value        

    @property
    def idattribute(self): 
        return self._idattribute
    @idattribute.setter
    def idattribute(self, value):
        self._idattribute = value     

    @property
    def idcfgname(self): 
        return self._idcfgname
    @idcfgname.setter
    def idcfgname(self, value):
        self._idcfgname = value  

    @property
    def pdmweid(self): 
        return self._pdmweid
    @pdmweid.setter
    def pdmweid(self, value):
        self._pdmweid = value        


    #endregion

    #region "ctor"
    def __init__(self,  **kwargs): 

        #region "Property defaults"
        self._configuration = {}
        self._aliasset = ""
        self._pdmweid = ""     
        self._id = ""
        self._idattribute = ""
        self._idcfgname = ""

        #endregion

        #region "Set Meta info"
        SerialBase.__init__(self , SerialF(fname="Document", rt=3) , **kwargs)           

        SerialT(self, "rt")
        SerialT(self, "aliasset" , pCol="TEXT1" )
        SerialT(self, "pdmweid" , pCol="TEXT2" )
        SerialT(self, "id" , pCol="TEXT3" )
        SerialT(self, "idattribute" , pCol="TEXT4" )
        SerialT(self, "aliasset" , pCol="TEXT5" )
        SerialT(self, "idcfgname" , pCol="TEXT6" )

        #endregion
    
    #endregion

class xmlTransactionsTransactionDocumentConfiguration(SerialBase) :
    
    #region Properties
    @property
    def attribute(self): 
        return self._attribute
    @attribute.setter
    def attribute(self, value):    
        self._attribute = []   
        for i in range(len(value)):
            try:
                self._attribute.append(xmlTransactionsTransactionDocumentConfigurationAttribute(**value[i]))
            except:
                self._attribute.append(xmlTransactionsTransactionDocumentConfigurationAttribute(**value))

    @property
    def references(self): 
        return self._references
    @references.setter
    def references(self, value):
        self._references = []
        for i in range(len(value["document"])):
            self._references.append(xmlTransactionsTransactionDocument(**value['document'][i]))

    @property
    def name(self): 
        return self._name
    @name.setter
    def name(self, value):
        self._name = value 

    @property
    def quantity(self): 
        return self._quantity
    @quantity.setter
    def quantity(self, value):
        self._quantity = value 

    #endregion

    #region "ctor"
    def __init__(self,  **kwargs): 

        #region "Property defaults"
        self._attribute = []
        self._references = []
        self._name = ""
        self._quantity = 0.0
        
        #endregion

        #region "Set Meta info"
        SerialBase.__init__(self , SerialF(fname="Configuration", rt=4) , **kwargs)   

        SerialT(self, "rt")        
        SerialT(self, "name" , pCol="TEXT1" )
        SerialT(self, "quantity" , pCol="REAL1" , pType="REAL" )

        #endregion
    
    #endregion
    
    #region Methods
    def byName(self, name):
        for a in self._attribute:
            if name.upper() == a.name.upper(): 
                return a.value
        return ''

    def kwargs(self):
        kw = {}
        for a in self._attribute:
            kw[a.name.replace(" ", "_").replace("/","").replace("\\" , "")] = a.value
        return kw

    #endregion
        
class xmlTransactionsTransactionDocumentConfigurationAttribute(SerialBase) :

    #region Properties
    @property
    def name(self): 
        return self._name
    @name.setter
    def name(self, value):
        self._name = value    
    
    @property
    def value(self): 
        return self._value
    @value.setter
    def value(self, value):
        self._value = value   
        
    #endregion 

    #region "ctor"
    def __init__(self,  **kwargs): 

        #region "Property defaults"
        self._name = ''        
        self._value = ''     
        
        #endregion

        #region "Set Meta info"
        SerialBase.__init__(self , SerialF(fname="Attribute", rt=5) , **kwargs)  

        SerialT(self, "rt")         
        SerialT(self, "name" , pCol="TEXT1" )
        SerialT(self, "value" , pCol="TEXT2" )

        #endregion
    
    #endregion



######################################################
## si@medatechuk.com
## 30/10/21
## https://github.com/MedatechUK/Medatech.APY
## 
## Logging class
## https://docs.python.org/3/howto/logging.html
##
## Usage: Start the Logger: !!USE ONCE!! in entry file
## from MedatechUK.mLog import mLog
##
##      # Creates a log reference to the running file
##        log = mLog()
##      # Start the Logger: !!USE ONCE!! in entry file
##        log.start( os.path.dirname(__file__), "DEBUG" )
##
## Usage: 
## from MedatechUK.mLog import mLog
##
##      # Creates a log reference to the running file
##        log = mLog()
##      # Write to the log
##        log.logger.info("Hello test!")

import logging
import inspect
from datetime import datetime
import os
from os.path import exists
import shutil
from pathlib import Path

class mLog():

    ## Ctor
    def __init__(self):  
        
        ## Build stack trace of caller frame             
        tree = [] 
        previous_frame = inspect.currentframe().f_back
        (filename) = inspect.getframeinfo(previous_frame)        
        tree.append(os.path.basename(filename.filename))

        previous_frame = previous_frame.f_back        
        while previous_frame and 'runpy.py' not in os.path.basename(filename.filename) and 'win32serviceutil.py' not in os.path.basename(filename.filename):
            (filename) = inspect.getframeinfo(previous_frame)        
            if os.path.basename(filename.filename) != tree[-1] and 'runpy.py' not in os.path.basename(filename.filename) and 'win32serviceutil.py' not in os.path.basename(filename.filename):
                tree.append(os.path.basename(filename.filename))
            previous_frame = previous_frame.f_back
        
        t = tree[-1]
        for i in range(len(tree)-2, -1 , -1):
            t = "{} > {}".format(t , tree[i] )

        ## Start the logger with stack trace %(name)s
        self.logger = logging.getLogger(t)

    def start(self, path , level):        
        ## Set the Log location
        now = datetime.now() # current date and time
        fn = '{}\log\{}-{}\{}.log'.format(
                path , 
                now.strftime("%Y") , 
                now.strftime("%m"), 
                now.strftime("%y%m%d")
            )
        try:
            os.makedirs(os.path.dirname(fn),exist_ok=True)
        except OSError as e:
            raise
            
        ## Set the configuration for the log
        logging.basicConfig(
            filename= fn, 
            encoding='utf-8', 
            format='%(asctime)s %(levelname)s %(name)s %(message)s', datefmt='%H:%M:%S',
            level=getattr(logging, level.upper())
        )        

    def logFile(self , f):
        if not exists(f):
            self.logger.critical("Failed to log file: {}".format(f))

        else:
            now = datetime.now() # current date and time
            try:
                fn = '{}\\{}\\'.format(
                    os.path.dirname(logging.root.handlers[0].baseFilename).rstrip("\\") ,  
                    now.strftime("%d")
                )            
                os.makedirs(os.path.dirname(fn),exist_ok=True)

                fi = f.split("\\")[-1].split(".")[0]
                ex = f.split("\\")[-1].split(".")[1]
                newf = "{}.{}".format(fi , ex)

                if exists("{}\\{}".format(fn , newf)):
                    i = 1
                    while exists("{}\\{}-{}.{}".format(fn , fi , i , ex)):
                        i += 1
                
                    newf = "{}-{}.{}".format(fi , i , ex)

                self.logger.info("Moving file from {} to {}".format(f , "{}\\".format(fn)))
                shutil.move(f, "{}\\{}".format(os.path.dirname(f).rstrip("\\"), newf))
                shutil.move("{}\\{}".format(os.path.dirname(f).rstrip("\\") , newf), "{}\\".format(fn) )

            except OSError as e:
                self.logger.critical("Error logging file: {}".format(str(e)))                



######################################################
## si@medatechuk.com
## 12/09/21
## https://github.com/MedatechUK/Medatech.APY
##
## Load configuration file containg settings for the
## loading. This may be either the IIS web.config or
## from a file called constants.py in the root dir.
##
## Example constant.py:
#    oDataHost ="walrus.ntsa.uk"
#    tabulaini ="tabula.ini"
#    ouser ="apiuser"
#    opass ="123456"
#    Environment = "wlnd"

import os, json
import xmltodict
import inspect
import pyodbc
import configparser

from MedatechUK.mLog import mLog

class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self

class Config:

    ## Ctor
    def __init__(self, **kwargs):

        ## Register for log
        self.log = mLog()

        ## Init Vars
        self.config = {}
        self.oDataHost = ''
        self.tabulaini = ''
        self.ouser = ''
        self.opass = ''
        self.path = ""
        self.connstr = 'Server=localhost\PRI,1433;Trusted_Connection=Yes;MultipleActiveResultSets=true;'

        for arg in kwargs.keys():
            ## If it's a request then take the environment
            #  from the request object
            if arg == 'request' :
                self.path = kwargs[arg].path
                self.environment = kwargs[arg].environment

                ## Try the web.config first
                if os.path.isfile(self.path + "\web.config"):
                    self.SettingfromWebConfig()
                else:
                    raise ("Not found [{}\\web.config].".format( self.path ))

            ## If there's no request (i.e. not a web integration)
            #  use the odata environment
            if arg == 'env':
                self.path = kwargs['path']
                self.environment = kwargs[arg]
                self.settingsini = self.path + "\\settings.ini"

                #region "Verify .ini"
                save = False
                if os.path.isfile(self.settingsini):
                    config = configparser.ConfigParser()
                    config.read(self.settingsini)
                    if 'odata' not in config.sections():
                        config['odata'] = {}
                        save = True
                    if not config['odata'].__contains__("odatahost"):
                        config['odata']['oDataHost'] = 'priority.localhost'
                        save = True
                    if not config['odata'].__contains__("tabulaini"):
                        config['odata']['tabulaini'] = 'tabula.ini'
                        save = True
                    if not config['odata'].__contains__("ouser"):
                        config['odata']['ouser'] = 'apiuser'
                        save = True
                    if not config['odata'].__contains__("opass"):
                        config['odata']['opass'] = 'password'
                        save = True
                    if not config['odata'].__contains__("ENV"):
                        config['odata']['ENV'] = 'demo'
                        save = True

                    if 'db' not in config.sections():
                        config['db'] = {}
                        save = True
                    if not config['db'].__contains__("server"):
                        config['db']['SERVER'] = 'localhost\PRI'
                        save = True
                    if not config['db'].__contains__("credentials"):
                        config['db']['CREDENTIALS'] = 'Trusted_Connection=Yes'
                        save = True

                else:
                    save = True
                    config = configparser.ConfigParser()
                    config['odata'] = {}
                    config['odata']['oDataHost'] = 'priority.localhost'
                    config['odata']['tabulaini'] = 'tabula.ini'
                    config['odata']['ouser'] = 'apiuser'
                    config['odata']['opass'] = 'password'
                    config['odata']['ENV'] = 'demo'
                    config['db'] = {}
                    config['db']['SERVER'] = 'localhost\PRI'
                    config['db']['CREDENTIALS'] = 'Trusted_Connection=Yes'
                    with open(self.settingsini, 'w') as configfile:
                        config.write(configfile)

                if save :
                    with open(self.settingsini, 'w') as configfile:
                        config.write(configfile)
                    raise NameError("Made [{}]: Please config.".format( self.settingsini ))

                #endregion
                 
                self.SettingfromConstants()

    ## Load setting from the IIS web.config
    def SettingfromWebConfig(self):
        ## Load the config file
        self.log.logger.debug("Opening [{}].".format( self.path + '\web.config' ))
        with open(self.path + '\web.config') as fd:
            self.config = xmltodict.parse(fd.read(), process_namespaces=True)

        ## Get the oData settings from the web.config
        for k in self.config['configuration']['appSettings']['add']:
            match k['@key'].upper():
                case "ODATAHOST": self.oDataHost = k['@value'].split("//")[1]
                case 'TABULAINI': self.tabulaini = k['@value']
                case 'OUSER': self.ouser = k['@value']
                case 'OPASS': self.opass = k['@value']

        ## Get the Priority Database connection string from the web.config
        if str(type(self.config['configuration']['connectionStrings']['add'])) =="<class 'list'>":
            for k in self.config['configuration']['connectionStrings']['add']:
                if k['@name'].upper() == 'PRIORITY':
                    self.connstr = k['@connectionString']
        if str(type(self.config['configuration']['connectionStrings']['add'])) =="<class 'collections.OrderedDict'>":
            if self.config['configuration']['connectionStrings']['add']['@name'].upper() == 'PRIORITY':
                self.connstr = self.config['configuration']['connectionStrings']['add']['@connectionString']

    def SettingfromConstants(self):
        ## Load settings from file settings.ini
        self.log.logger.debug("Opening [{}].".format( self.path + '\settings.ini' ))
        config = configparser.ConfigParser(dict_type=AttrDict)
        config.read(self.path + '//settings.ini')

        self.oDataHost = config._sections.odata.odatahost
        self.tabulaini =  config._sections.odata.tabulaini
        self.ouser =  config._sections.odata.ouser
        self.opass =  config._sections.odata.opass
        self.connstr = "SERVER={};{};MARS_Connection=Yes".format(
            config._sections.db.server
            , config._sections.db.credentials
        )
        self.config = config._sections

    def CheckEnviroment(self):
        self.log.logger.info("Checking environment [{}].".format( self.environment ))
        try:
            cnxn = self.cnxn()
        except:
            return False

        crsr = cnxn.cursor()
        crsr.execute("select DNAME from ENVIRONMENT where DNAME <> '' union all select 'system'")
        for row in [row for row in crsr.fetchall() if row.DNAME.lower() == self.environment.lower()] :
            # Set the environment to the cAsE of the db object
            self.environment = row.DNAME
            self.log.logger.info("Environment [{}] OK!".format( self.environment ))
            return True

        # Environment not found
        return False

    def cnxn(self):
        cnstr = ''
        # Find installed driver
        for dr in [dr for dr in pyodbc.drivers() if "for SQL Server" in dr]:
            cnstr = "Driver={{{}}};DATABASE=system;{}".format( dr , self.connstr )
            try:
                return pyodbc.connect(cnstr)
            except Exception as e:
                self.log.logger.critical("Could not connect to {}.".format(cnstr))
                raise

        self.log.logger.critical("No pyodbc driver installed.")
        raise TypeError("No pyodbc driver")

                  